/************************************************************************
 *  emulation_c.c
 *                      Main emulation routines for PalmBoy
 *                      (C functions)
 *
 *  2000-05-01  Bodo Wenzel   Got from full source v.0.4 as runmode.asm
 *  2000-05-06  Bodo Wenzel   First working version with built-in games
 *  2000-05-07  Bodo Wenzel   First working version
 *  2000-08-02  Bodo Wenzel   Now nearly full screen emulation
 *  2000-09-13  Bodo Wenzel   Some things in separated files
 *  2000-09-30  Bodo Wenzel   Scroller in GB code
 *  2000-10-22  Bodo Wenzel   Interrupts and timer
 *  2000-11-03  Bodo Wenzel   Request of EasterEgg corrected
 *  2000-11-21  Bodo Wenzel   Support of color devices and
 *                            better joypad emulation
 *  2001-03-18  Bodo Wenzel   Support of more cartridges,
 *                            new stack management, STOP corrected
 *  2001-06-23  Bodo Wenzel   v.3.3, Bug in MBC1 code corrected
 *  2001-07-01  Bodo Wenzel   v.3.3b, Bugfix corrected ;-)
 *  2001-07-27  Bodo Wenzel   v.4.0beta, changed to C
 *  2001-09-16  Bodo Wenzel   Saving states now possible
 * wenn spaeter mehrere States gespeichert werden, muss natuerlich auch
 * das Cartridge-RAM mitgesichert werden - und die Spielzeit usw.!
 ************************************************************************
 *
 * (c)2000-2001 Bodo Wenzel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

/* IMPORTANT!
 * Please make sure that any referenced resource is included!
 * Unfortunately there is no automatic!
 */

/* === Includes =======================================================	*/

#include <PalmOS.h>

#include "palmboy.h"
#include "emulation.h"
#include "list.h"
#include "video.h"
#include "memory.h"
#include "gbemu.h"

/* === Constants ======================================================	*/

#define rsrcIdEmulator (2)

/* NOTE:
 * The databases are created on card #0. For the moment I'll do no effort
 * to change this behaviour.
 */
#define dbCard (0)

#define savedStateVersion (0x0009) /* major, minor version */

#define stateDbType 'DATA'

#define lineBegX (0)   /* coordinates for separator */
#define lineEndX (159)
#define linePosY (144)

/* Key bits are inverted: released/pressed = 1/0 */
#define gbCursorMask  (0xef) /* bits: 1 1 1 0 down up left right */
#define gbControlMask (0xdf) /* bits: 1 1 0 1 start select b a */

#define myKeys    (keyBitPageUp|keyBitPageDown| \
                   keyBitHard1|keyBitHard2|keyBitHard3|keyBitHard4)
#define myKeyMask (keyBitsAll&~myKeys)

/* === Type definitions ===============================================	*/

typedef struct {
  UInt16             version;
  EmulationStateType emulation;
  VideoStateType     video;
  MemoryStateType    memory;
} SavedStateType;

/* === Global and static variables ====================================	*/

Boolean EmulationRuns; /* there's a state database */

static DmOpenRef StateDbP = NULL;
static Boolean SaveState;
static SavedStateType *SavedStateP = NULL;
static Char StateDbName[] = "PalmBoy-State";
static Char CrashDbName[] = "PalmBoy-Crash";

static MemHandle EmulatorH;

static UInt16 GameCardNo;
static Char GameName[dmDBNameLength];
static DmOpenRef GameDbP = NULL;

static FormType *FormP; /* form variables */
static UInt16 SelectButton;
static RectangleType SelectBounds;
static UInt16 StartButton;
static RectangleType StartBounds;

/* Address of emulator kernel; NOTE: This points to combo 0x00 at offset
 * 0x8000 because of SInt16 indices!
 */
GbemuComboType *EmulationEmulator = NULL;

UInt8 EmulationIntIme; /* interrupt master enable */
UInt8 EmulationIntIm;  /* mask = IME & IE */

UInt8 EmulationError; /* error code from emulator */

UInt8 *EmulationJoypadPtr;
UInt8 EmulationJoypad[4];
UInt8 EmulationJoypadOld; /* saved joypad value before user poll */

/* === Function prototypes ============================================	*/

static Boolean SetupAll(void);
static void CleanupAll(void);
static void StartEmulation(void);

static void EmulationLoadState(EmulationStateType *stateP, Boolean reset);
static void EmulationSaveState(void *recP, UInt32 offset);
static void EmulationSetup(void);

/* === Functions for the emulator kernel ==============================	*/

Boolean EmulationInit(void) {
  /*-----------------------*/
  /* just check and reserve the resource of the emulator kernel */
  EmulationRuns = DmFindDatabase(dbCard, StateDbName) != 0;

  EmulatorH = DmGet1Resource('code', rsrcIdEmulator);
  return EmulatorH != NULL;
}

void EmulationExit(void) {
  /*--------------------*/
  if (EmulatorH != NULL)
    DmReleaseResource(EmulatorH);
}

/* === User interface =================================================	*/

void EmulationRunGame(UInt16 cardNo, const Char *nameP) {
  /*---------------------------------------------------*/
  GameCardNo = cardNo;
  StrCopy(GameName, nameP);

  FrmGotoForm(formIdEmulation);
}

Boolean EmulationFormHandleEvent(EventPtr evtP) {
  /*-------------------------------------------*/
  UInt32 keyOldMask;

  /* we're waiting for the form to open */
  if (evtP->eType != frmOpenEvent)
    return false;

  if (!SetupAll())
    FrmAlert(alertIdSetupError);
  else {
    keyOldMask = KeySetMask(myKeyMask);

    FormP = FrmGetActiveForm();
    FrmDrawForm(FormP);

    WinDrawLine(lineBegX, linePosY, lineEndX, linePosY);

    SelectButton = FrmGetObjectIndex(FormP, buttonIdSelect);
    FrmGetObjectBounds(FormP, SelectButton, &SelectBounds);
    StartButton = FrmGetObjectIndex(FormP, buttonIdStart);
    FrmGetObjectBounds(FormP, StartButton, &StartBounds);

    StartEmulation();

    KeySetMask(keyOldMask);
  }

  CleanupAll();

  EmulationRuns = false;
  FrmGotoForm(formIdList);
  return true;
}

Boolean EmulationUserPoll(void) {
  /*---------------------------*/
  UInt8  jpCursor, jpControl;
  UInt32 keys;

  MemSemaphoreRelease(true);

  jpCursor = 0xef;
  jpControl = EmulationJoypad[emulationJoypadCtrl] | 0x03;

  keys = KeyCurrentState();
  /* bits: 1 1 1 0 down up left right */
  if (keys & keyBitPageUp)
    jpCursor &= ~0x04;
  if (keys & keyBitPageDown)
    jpCursor &= ~0x08;
  if (keys & keyBitHard1)
    jpCursor &= ~0x02;
  if (keys & keyBitHard2)
    jpCursor &= ~0x01;
  /* bits: 1 1 0 1 start select b a */
  if (keys & keyBitHard3)
    jpControl &= ~0x02;
  if (keys & keyBitHard4)
    jpControl &= ~0x01;

  if (EvtEventAvail()) {
    EventType evt;

    EvtGetEvent(&evt, 0);
    switch (evt.eType) {
    case nilEvent:
      break;

    case penDownEvent:
    case penMoveEvent:
    {
      UInt16 x, y;

      x = evt.screenX;
      y = evt.screenY;
      if (y < 160) {
	WinDisplayToWindowPt(&x, &y);
	if (RctPtInRectangle(x, y, &SelectBounds)) {
	  jpControl &= ~0x04;
	  FrmSetControlValue(FormP, SelectButton, true);
	}
	if (RctPtInRectangle(x, y, &StartBounds)) {
	  jpControl &= ~0x08;
	  FrmSetControlValue(FormP, StartButton, true);
	}
      } else
	SysHandleEvent(&evt);
      break;
    }

    case penUpEvent:
      jpControl |= 0x04 | 0x08;
      FrmSetControlValue(FormP, SelectButton, false);
      FrmSetControlValue(FormP, StartButton, false);
      if (evt.screenY >= 160)
	SysHandleEvent(&evt);
      break;

    case keyDownEvent:
      if ((evt.data.keyDown.modifiers & commandKeyMask) == 0) {
	SysHandleEvent(&evt);
	break;
      }

      switch (evt.data.keyDown.chr) {
      case menuChr: /* special meaning for "Menu" */
	SndPlaySystemSound(sndClick);
	MemSemaphoreReserve(true);
	return true;
      default:
	SysHandleEvent(&evt);
	break;
      }
      break;

    case appStopEvent:
      EvtAddEventToQueue(&evt);
      SndPlaySystemSound(sndClick);
      SaveState = true;
      MemSemaphoreReserve(true);
      return true;

    default:
      SysHandleEvent(&evt);
      break;
    }
  }

  MemSemaphoreReserve(true);

  EmulationJoypad[emulationJoypadCurs] = jpCursor;
  EmulationJoypad[emulationJoypadCtrl] = jpControl;
  EmulationJoypad[emulationJoypadBoth] = jpCursor & jpControl;

  return false;
}

/* === Running a game =================================================	*/

static Boolean SetupAll(void) {
  /*-------------------------*/
  LocalID dbId;
  Boolean reset;

  /* prevent saving a state */
  SaveState = false;
  EmulationError = gbemuRetUserStop;

  /* lock the resource with the emulator kernel */
  EmulationEmulator = (GbemuComboType *)MemHandleLock(EmulatorH) + 0x80;
  if (EmulationEmulator == NULL)
    return false;

  /* try to open the state database, on errors create new one */
  dbId = DmFindDatabase(dbCard, StateDbName);
  if (dbId != 0) {
    StateDbP = DmOpenDatabase(dbCard, dbId, dmModeReadWrite);
    if (StateDbP != NULL) {
      if (DmNumRecords(StateDbP) == 1) {
	MemHandle recH;

	recH = DmGetRecord(StateDbP, 0);
	if (recH != NULL) {
	  if (MemHandleSize(recH) == sizeof(SavedStateType)) {
	    SavedStateP = MemHandleLock(recH);
	    if (SavedStateP != NULL) {
	      if (SavedStateP->version != savedStateVersion) {
		MemHandleUnlock(recH);
		SavedStateP = NULL;
	      }
	    }
	  }
	  if (SavedStateP == NULL)
	    DmReleaseRecord(StateDbP, 0, false);
	}
      }
      if (SavedStateP == NULL) {
	DmCloseDatabase(StateDbP);
	StateDbP = NULL;
      }
    }
  }

  if (StateDbP != NULL)
    reset = false;
  else {
    /* no or wrong state database, create new one */
    UInt16 recIndex;
    MemHandle recH;

    reset = true;

    dbId = DmFindDatabase(dbCard, StateDbName);
    if (dbId != 0)
      DmDeleteDatabase(dbCard, dbId);
    if (DmCreateDatabase(dbCard, StateDbName, creator,
			 stateDbType, false) != errNone)
      return false;
    dbId = DmFindDatabase(dbCard, StateDbName);
    if (dbId == 0)
      return false;
    StateDbP = DmOpenDatabase(dbCard, dbId, dmModeReadWrite);
    if (StateDbP == NULL)
      return false;

    recIndex = 0;
    recH = DmNewRecord(StateDbP, &recIndex, sizeof(SavedStateType));
    if (recH != NULL) {
      SavedStateP = MemHandleLock(recH);
      if (SavedStateP != NULL) {
	UInt16 version;

	/* make sure that it will be rejected until filled correctly */
	version = 0;
	DmWrite(SavedStateP, OffsetOf(SavedStateType, version),
		&version, sizeof(version));
      } else
	DmReleaseRecord(StateDbP, 0, false);
    }
    if (SavedStateP == NULL)
      return false;
  }

  EmulationLoadState(&(SavedStateP->emulation), reset);
  VideoLoadState(&(SavedStateP->video), reset);
  MemoryLoadState(&(SavedStateP->memory), reset);

  /* try to open the game's database, and last setup's */
  dbId = DmFindDatabase(GameCardNo, GameName);
  if (dbId == 0)
    return false;
  GameDbP = DmOpenDatabase(GameCardNo, dbId, dmModeReadWrite);
  if (GameDbP == NULL)
    return false;
  if (!MemorySetup(GameDbP))
    return false;
  EmulationSetup();
  if (!VideoSetup(StateDbP, &(SavedStateP->video)))
    return false;
  return true;
}

static void CleanupAll(void) {
  /*------------------------*/
  LocalID dbId;

  /* an emulation error occurred, setup for a crash-database */
  if (EmulationError != gbemuRetUserStop) {
    FrmAlert(alertIdEmulationError);

    dbId = DmFindDatabase(dbCard, CrashDbName);
    if (dbId != 0)
      DmDeleteDatabase(dbCard, dbId);

    DmOpenDatabaseInfo(StateDbP, &dbId, NULL, NULL, NULL, NULL);
    DmSetDatabaseInfo(dbCard, dbId, CrashDbName, NULL, NULL, NULL, NULL,
		      NULL, NULL, NULL, NULL, NULL, NULL);
    SaveState = true;
  }

  /* release the game's database */
  if (GameDbP != NULL) {
    MemoryCleanup(GameDbP);
    DmCloseDatabase(GameDbP);
    GameDbP = NULL;
  }

  /* release state database, remove scratch memory,
   * save state, if appropriate
   */
  if (StateDbP != NULL) {
    if (SavedStateP != NULL) {
      if (SaveState) {
	UInt16 version;

	EmulationSaveState(SavedStateP,
			   OffsetOf(SavedStateType, emulation));
	VideoSaveState(SavedStateP, OffsetOf(SavedStateType, video));
	MemorySaveState(SavedStateP, OffsetOf(SavedStateType, memory));
	version = savedStateVersion;
	DmWrite(SavedStateP, OffsetOf(SavedStateType, version),
		&version, sizeof(version));
      }
      MemPtrUnlock(SavedStateP);
      DmReleaseRecord(StateDbP, 0, true);
      SavedStateP = NULL;
      VideoCleanup();
      {
	UInt16 i;

	for (i = DmNumRecords(StateDbP) - 1; i > 0; i--) {
	  DmReleaseRecord(StateDbP, i, false);
	  DmRemoveRecord(StateDbP, i);
	}
      }
    }
    DmCloseDatabase(StateDbP);
    StateDbP = NULL;

    if (!SaveState) {
      dbId = DmFindDatabase(dbCard, StateDbName);
      if (dbId != 0)
	DmDeleteDatabase(dbCard, dbId);
    }
  }

  /* release the resource with the emulator kernel */
  if (EmulationEmulator != NULL) {
    MemHandleUnlock(EmulatorH);
    EmulationEmulator = NULL;
  }
}

static void StartEmulation(void) {
  /*----------------------------*/
  EmulationCpuStateType save;

  MemSemaphoreReserve(true);

  /* in case a game is continued, save the current CPU state */
  save = EmulationCpuState;

  /* start restore code in GB machine language */
  MemoryBaseRom0 = EmulationEmulator[gbemuGbCodeReset].xt.code;
  EmulationCpuState.pcBase = EmulationEmulator[gbemuGbCodeReset].xt.code;
  EmulationCpuState.pcOffset = 0x0000;
  EmulationCpuState.a = EmulationRuns;
  EmulationCpuState.b = EmulationIo.p1;
  EmulationCpuState.c = EmulationIo.stat;
  EmulationCpuState.d = VideoSavedLcdc;
  EmulationCpuState.e = PrefGetPreference(prefAllowEasterEggs) != 0 ?
    0xff : 0x00;

  /* give enough cycles for not calling any VideoLineFunction too early */
  VideoCycles = 64000;
  EmulationStartKernel();
  /* don't save if stopped in restart code... */
  SaveState = false;

  /* the GB setup code ends with an unknown opcode,
   * or an error occurred, or the user stopped
   */
  if (EmulationError == gbemuRetOpcodeUnknown) {
    /* show current screen in any case */
    VideoCopyScreen();
#if 0 /* debug */
    MemSemaphoreRelease(true);
    SysTaskDelay(100);
    MemSemaphoreReserve(true);
#endif

    /* Start with video support immediately after restoring PC and SP.
     * Look carefully in gbemu_xt.inc for the count of needed cycles...
     */
    MemoryBaseRom0 = MemoryRomStart;
    if (!EmulationRuns)
      EmulationCpuState.pcOffset = 0x0100;
    else
      EmulationCpuState = save;
    EmulationEmulator[gbemuGbCodeRestart].xt.code[0x01] =
      EmulationCpuState.spOffset & 0xff;
    EmulationEmulator[gbemuGbCodeRestart].xt.code[0x02] =
      EmulationCpuState.spOffset >> 8;
    EmulationEmulator[gbemuGbCodeRestart].xt.code[0x04] =
      EmulationCpuState.pcOffset & 0xff;
    EmulationEmulator[gbemuGbCodeRestart].xt.code[0x05] =
      EmulationCpuState.pcOffset >> 8;
    EmulationCpuState.pcBase =
      EmulationEmulator[gbemuGbCodeRestart].xt.code;
    EmulationCpuState.pcOffset = 0x0000;
    VideoCycles = 2;
    EmulationEmulator[gbemuJmpLcdLine].xt.vector.address =
      VideoUserPoll;
    EmulationStartKernel();
  }

  MemSemaphoreRelease(true);
}

/* === Load, save, and setup state ====================================	*/

static void EmulationLoadState(EmulationStateType *stateP, Boolean reset) {
  /*---------------------------------------------------------------------*/
  if (reset) {
    EmulationIntIme = 0;
    EmulationIo.if_ = 0x00;

  } else {
    GameCardNo = stateP->cardNo;
    StrNCopy(GameName, stateP->name, sizeof(GameName));

    EmulationIntIme = stateP->intIme;
    EmulationCpuState = stateP->cpuState;
    EmulationIo = stateP->io;
  }

  EmulationIo.sb = 0xff;
  EmulationIo.dma = 0xff;
}

static void EmulationSaveState(void *recP, UInt32 offset) {
  /*-----------------------------------------------------*/
  DmWrite(recP, offset + OffsetOf(EmulationStateType, cardNo),
	  &GameCardNo, sizeof(GameCardNo));
  DmWrite(recP, offset + OffsetOf(EmulationStateType, name),
	  &GameName, sizeof(GameName));

  DmWrite(recP, offset + OffsetOf(EmulationStateType, intIme),
	  &EmulationIntIme, sizeof(EmulationIntIme));
  DmWrite(recP, offset + OffsetOf(EmulationStateType, cpuState),
	  &EmulationCpuState, sizeof(EmulationCpuState));

  EmulationIo.p1 = (EmulationJoypadPtr - EmulationJoypad) << 4;
  DmWrite(recP, offset + OffsetOf(EmulationStateType, io),
	  &EmulationIo, sizeof(EmulationIo));
}

static void EmulationSetup(void) {
  /*----------------------------*/
  /* set vectors */
  MemSemaphoreReserve(true);
  EmulationEmulator[gbemuJmpOpcodeStop].xt.vector.address =
    EmulationOpcodeStop;
  MemSemaphoreRelease(true);

  EmulationJoypad[emulationJoypadBoth] = gbCursorMask & gbControlMask;
  EmulationJoypad[emulationJoypadCtrl] = gbControlMask;
  EmulationJoypad[emulationJoypadCurs] = gbCursorMask;
  EmulationJoypad[emulationJoypadNone] = 0xff;
  EmulationJoypadOld = 0xff;
}

/* === The end ========================================================	*/
