/************************************************************************
 *  list.c
 *                      List form routines for PalmBoy
 *
 *  2000-05-01  Bodo Wenzel   Got from full source v.0.4
 *  2000-05-06  Bodo Wenzel   First working version with built-in games
 *  2000-05-07  Bodo Wenzel   First working version
 *  2000-11-03  Bodo Wenzel   Removed bug if no games present
 *  2000-11-14  Bodo Wenzel   Support of color devices
 *  2001-02-18  Bodo Wenzel   Support of more cartridges
 *  2001-07-23  Bodo Wenzel   Changed to C
 *  2001-09-16  Bodo Wenzel   Saving states now possible
 ************************************************************************
 *
 * (c)2000-2001 Bodo Wenzel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

/* IMPORTANT!
 * Please make sure that any referenced resource is included!
 * Unfortunately there is no automatic!
 */

/* === Includes =======================================================	*/

#include <PalmOS.h>

#include "palmboy.h"
#include "list.h"
#include "emulation.h"

/* === Constants ======================================================	*/

#define gameType ('GAME')

/* === Global variables ===============================================	*/

static SysDBListItemType *DbList;
static UInt16            AllRows, VisibleRows, TopRow;

/* === Function prototypes ============================================	*/

static void InfoShow(UInt16 no);
static Boolean DeleteGame(UInt16 no);
static void Update(FormPtr frmP, TablePtr tblP);
static SysDBListItemType *GetItem(UInt16 no);
static void DrawOneTitle(void *tableP, Int16 row, Int16 col,
			 RectanglePtr bounds);

/* === Functions ======================================================	*/

void ListInit(void) {
  /*---------------*/
  FormPtr   frmP;
  UInt16    dbCount;
  MemHandle listH;

  AllRows = 0;
  DbList = NULL;

  frmP = FrmInitForm(formIdListSummarize);
  FrmDrawForm(frmP);

  if (SysCreateDataBaseList(gameType, creator, &dbCount, &listH, true)) {
    AllRows = dbCount;
    DbList = MemHandleLock(listH);
  }

  TopRow = 0;
  FrmEraseForm(frmP);
}

void ListExit(void) {
  /*---------------*/
  MemHandle listH;

  if (DbList != NULL) {
    listH = MemPtrRecoverHandle(DbList);
    MemHandleUnlock(listH);
    MemHandleFree(listH);
    DbList = NULL;
  }
}

Boolean ListFormHandleEvent(EventPtr evtP) {
  /*--------------------------------------*/
  FormPtr           frmP;
  TablePtr          tblP;
  ScrollBarPtr      sclP;
  Int16             row;
  SysDBListItemType *itemP;
  Boolean           ret;
  Int16             dummy, min, max, size;

  frmP = FrmGetActiveForm();
  tblP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, tableIdList));

  switch (evtP->eType) {

  case frmOpenEvent:
    VisibleRows = TblGetNumberOfRows(tblP);
    for (row = 0; row < VisibleRows; row++)
      TblSetItemStyle(tblP, row, 0, customTableItem);
    TblSetCustomDrawProcedure(tblP, 0, DrawOneTitle);
    TblSetColumnUsable(tblP, 0, true);

    Update(frmP, tblP);

    FrmSetControlGroupSelection(frmP, groupIdList, pushbuttonIdRun);

    FrmDrawForm(frmP);
    return true;

  case tblSelectEvent:
    row = evtP->data.tblSelect.row + TopRow;

    switch (FrmGetObjectId(frmP,
			   FrmGetControlGroupSelection(frmP, groupIdList) &
			   0xFF)) {

    case pushbuttonIdRun:
      itemP = GetItem(row);
      if (itemP != NULL)
	EmulationRunGame(itemP->cardNo, itemP->name);
      break;

    case pushbuttonIdInfo:
      InfoShow(row);
      break;

    case pushbuttonIdDelete:
      ret = DeleteGame(row);
      TblUnhighlightSelection(tblP);
      if (ret)
	Update(frmP, tblP);
      return true;

    default:
      break;
    }
    TblUnhighlightSelection(tblP);
    return true;

  case sclRepeatEvent:
    TopRow = evtP->data.sclRepeat.newValue;
    TblMarkTableInvalid(tblP);
    TblRedrawTable(tblP);
    return false;

  case keyDownEvent:
    switch (evtP->data.keyDown.chr) {

    case pageUpChr:
      if (TopRow > VisibleRows)
        TopRow -= VisibleRows;
      else
        TopRow = 0;
      break;

    case pageDownChr:
      if (AllRows > VisibleRows) {
	TopRow += VisibleRows;
	if (TopRow > AllRows - VisibleRows)
	  TopRow = AllRows - VisibleRows;
      } else
	TopRow = 0;
      break;

    default:
      return false;
    }

    sclP = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, scrollbarIdList));
    SclGetScrollBar(sclP, &dummy, &min, &max, &size);
    SclSetScrollBar(sclP, TopRow, min, max, size);

    TblMarkTableInvalid(tblP);
    TblRedrawTable(tblP);
    return true;

  case menuEvent:
    if (evtP->data.menu.itemID == menuIdLOptionsAbout)
      FrmAlert(alertIdAbout);
    return true;

  default:
    return false;
  }
}

static void InfoShow(UInt16 no) {
  /*---------------------------*/
  SysDBListItemType  *itemP;
  LocalID            dbId;
  DmOpenRef          dbP;
  MemHandle          recordH;
  UInt8              *recordP;
  UInt16             romSize, ramSize;
  Char               sRom[4], sRam[4];
  static const UInt8 bLisRamSizes[] = { 0, 2, 8, 32 };

  itemP = GetItem(no);
  if (itemP == NULL)
    return;

  dbId = DmFindDatabase(itemP->cardNo, itemP->name);
  if (dbId == 0)
    return;
  dbP = DmOpenDatabase(itemP->cardNo, dbId, dmModeReadOnly);
  if (dbP == NULL)
    return;

  recordH = DmQueryRecord(dbP, 0);
  recordP = MemHandleLock(recordH);
  romSize = recordP[0x148];
  if (recordP[0x147] == 5 || recordP[0x147] == 6)
    ramSize = 1;
  else
    ramSize = bLisRamSizes[recordP[0x149] & 0x03];

  StrPrintF(sRam, "%3d", ramSize);
  StrPrintF(sRom, "%3d", 32 << romSize);
  FrmCustomAlert(alertIdInfo, itemP->name, sRom, sRam);

  MemHandleUnlock(recordH);
  DmCloseDatabase(dbP);
}

static Boolean DeleteGame(UInt16 no) {
  /*--------------------------------*/
  SysDBListItemType *itemP;
  LocalID           dbId;
  UInt16            rows;

  /* Here's a problem if the database is deleted, but not the entry.
   * Don't know, whether it's already gone...
   */

  itemP = GetItem(no);
  if (itemP == NULL)
    return false;

  if (FrmCustomAlert(alertIdDelete, itemP->name, NULL, NULL) != 0)
    return false;

  dbId = DmFindDatabase(itemP->cardNo, itemP->name);
  if (dbId != 0) {
    if (DmDeleteDatabase(itemP->cardNo, dbId) != errNone)
      return false; /* errors silently ignored */
  }

  if (--AllRows == 0)
    return true;

  rows = AllRows - no;
  if (rows == 0)
    return true;

  MemMove(itemP, itemP + 1, rows * sizeof(SysDBListItemType));
  return true;
}

static void Update(FormPtr frmP, TablePtr tblP) {
  /*-------------------------------------------*/
  ScrollBarPtr pScl;
  UInt16       maxRow, row;

  pScl = FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, scrollbarIdList));
  if (AllRows > VisibleRows)
    maxRow = AllRows - VisibleRows;
  else
    maxRow = 0;
  if (maxRow < TopRow)
    TopRow = maxRow;
  SclSetScrollBar(pScl, TopRow, 0, maxRow, VisibleRows);

  TblMarkTableInvalid(tblP);
  for (row = 0; row < VisibleRows; row ++)
    TblSetRowUsable(tblP, row, row < AllRows);
  TblRedrawTable(tblP);
}

static SysDBListItemType *GetItem(UInt16 no) {
  /*----------------------------------------*/
  if (AllRows == 0)
    return NULL;
  if (DbList == NULL)
    return NULL;
  return DbList + no;
}

static void DrawOneTitle(void *tableP, Int16 row, Int16 col,
			 RectanglePtr bounds) {
  /*------------------------------------------------------*/
  SysDBListItemType *itemP;

  itemP = GetItem(TopRow + row);
  if (itemP == NULL)
    return;

  WinDrawChars(itemP->name, StrLen(itemP->name),
	       bounds->topLeft.x, bounds->topLeft.y);
}

/* === The end ========================================================	*/
