/************************************************************************
 *  memory.h
 *                      Memory emulation routines for PalmBoy
 *
 *  2001-07-22  Bodo Wenzel   Created for v.4.0
 *  2001-09-16  Bodo Wenzel   Saving states now possible
 ************************************************************************
 *
 * (c)2000-2001 Bodo Wenzel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

#ifndef MEMORY_H
#define MEMORY_H

/* === Constants ======================================================	*/

#define memoryMaxRomPages (256)
#define memoryMaxRamPages (4)

/* === Global variables ===============================================	*/

#ifdef __GNUC__ /* only if C preprocessing */

typedef enum {
  mbcUnknown= 0, mbcDefault, mbc1, mbc2, mbc3, mbc5
} mbcType;

typedef struct {
  Int16      code;
  mbcType    mbc;
  const Char *name;
} MemoryCartridgeType;

extern const MemoryCartridgeType MemoryCartridgesKnown[];

typedef struct {
  UInt8 vid8000[0x2000]; /* video RAM */
  UInt8 intC000[0x2000]; /* internal RAM */
  UInt8 oamFe00[0x0100]; /* object attribute memory */
} MemoryRamType;

/* NOTE:
 * As GB addresses are used as offsets, the addresses for Palm memory
 * blocks are offset by the base addresses.
 */
extern UInt8 *MemoryRomStart;                    /* addr of 1st ROM page */
extern UInt8 *MemoryRomPages[memoryMaxRomPages]; /* addr's of ROM pages */
extern UInt8 *MemoryRamPages[memoryMaxRamPages]; /* addr's of RAM pages */

extern UInt8 *MemoryBaseRom0;
extern UInt8 *MemoryBaseRom1;
extern UInt8 *MemoryBaseRamVid;
extern UInt8 *MemoryBaseRamExt;
extern UInt8 *MemoryBaseRamInt;
extern UInt8 *MemoryBaseRamOam;

extern UInt16 MemoryMbc1RomLsb;
extern UInt16 MemoryMbc1RomMsb;
extern UInt8  MemoryMbc1Mode;

typedef struct {
  UInt8 seconds;
  UInt8 minutes;
  UInt8 hours;
  UInt8 days;
  UInt8 control;
} MemoryMbc3RtcType;

extern UInt16            MemoryMbc3RtcDivider; /* dividing frame rate */
extern MemoryMbc3RtcType MemoryMbc3RtcCurrent;
extern MemoryMbc3RtcType MemoryMbc3RtcLatched;
extern UInt8             MemoryMbc3Mode;
extern UInt8             *MemoryMbc3RtcPtr; /* ptr to current register */

/* types of variables of a state to be saved */

typedef struct {
  UInt16            romPage;
  UInt8             ramPage;
  UInt16            mbc1RomLsb;
  UInt16            mbc1RomMsb;
  UInt8             mbc1Mode;
  UInt8             mbc3Mode;
  UInt16            mbc3RtcDivider;
  MemoryMbc3RtcType mbc3RtcCurrent;
  MemoryMbc3RtcType mbc3RtcLatched;
} MemoryMbcStateType;

typedef struct {
  MemoryRamType      ram;
  MemoryMbcStateType mbc;
} MemoryStateType;

#endif

/* === Function prototypes in C part ==================================	*/

#ifdef __GNUC__ /* only if C preprocessing */

void MemoryLoadState(MemoryStateType *stateP, Boolean reset);
void MemorySaveState(void *recP, UInt32 offset);
Boolean MemorySetup(DmOpenRef dbP);
void MemoryCleanup(DmOpenRef dbP);

#endif

/* === Function prototypes in assembler part ==========================	*/

#ifdef __GNUC__ /* only if C preprocessing */

/* NOTE:
 * These functions are called from the emulation and get their parameters
 * via several registers. This is not documented here, the prototypes are
 * just for exporting the names.
 */

void MemoryMbcDummy(void);
void MemoryMbc1RomSelect(void);
void MemoryMbc1RomSelectL(void);
void MemoryMbc1RomSelectH(void);
void MemoryMbc1RamSelect(void);
void MemoryMbc1ModeSelect(void);
void MemoryMbc2RomSelect(void);
void MemoryMbc3RomSelect(void);
void MemoryMbc3RamSelect(void);
void MemoryMbc3RtcRd(void);
void MemoryMbc3RtcWr(void);
void MemoryMbc3RtcLatch(void);
void MemoryMbc3RtcWork(void);
void MemoryMbc5RomSelect(void);
void MemoryMbc5RamSelect(void);

void MemoryRamRdExtNone(void);
void MemoryRamWrExtNone(void);
void MemoryRamRdExtMbc2(void);
void MemoryRamWrExtMbc2(void);
void MemoryRamRdExt2K(void);
void MemoryRamWrExt2K(void);
void MemoryRamRdExt8K(void);
void MemoryRamWrExt8K(void);

void MemoryRamNone(void);
void MemoryRamTileBg(void);
void MemoryRamTileBgWi(void);
void MemoryRamTileOb(void);
void MemoryRamTileBgOb(void);
void MemoryRamTileBgWiOb(void);
void MemoryRamMapNone(void);
void MemoryRamMapBg(void);
void MemoryRamMapWi(void);
void MemoryRamMapBgWi(void);
void MemoryRamOam(void);
void MemoryRamOamEntry(void);

#endif

/* === The end ========================================================	*/

#endif
