/************************************************************************
 *  memory_c.c
 *                      Memory emulation routines for PalmBoy
 *                      (C functions)
 *
 *
 *  2000-09-10  Bodo Wenzel   Got from runmode.asm
 *  2000-09-17  Bodo Wenzel   Now nearly full screen emulation
 *  2000-10-19  Bodo Wenzel   Interrupts and timer
 *  2000-11-15  Bodo Wenzel   Support of color devices
 *  2001-02-24  Bodo Wenzel   Support of more cartridges
 *  2001-07-21  Bodo Wenzel   Separated from emumem.asm, changed to C
 *  2001-09-30  Bodo Wenzel   Saving states now possible
 ************************************************************************
 *
 * (c)2000-2001 Bodo Wenzel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

/* === Includes =======================================================	*/

#include <PalmOS.h>

#include "memory.h"
#include "emulation.h"
#include "gbemu.h"

/* === Global variables ===============================================	*/

const MemoryCartridgeType MemoryCartridgesKnown[] = {
  { 0x00, mbcDefault, "ROM ONLY" },
  { 0x01, mbc1,       "MBC1" },
  { 0x02, mbc1,       "MBC1+RAM" },
  { 0x03, mbc1,       "MBC1+RAM+BAT" },
  { 0x05, mbc2,       "MBC2" },
  { 0x06, mbc2,       "MBC2+BAT" },
  { 0x08, mbcDefault, "RAM" },
  { 0x09, mbcDefault, "RAM+BAT" },
  { 0x0b, mbcUnknown, "MMM01" },
  { 0x0c, mbcUnknown, "MMM01+RAM" },
  { 0x0d, mbcUnknown, "MMM01+RAM+BAT" },
  { 0x0f, mbc3,       "MBC3+BAT+RTC" },
  { 0x10, mbc3,       "MBC3+RAM+BAT+RTC" },
  { 0x11, mbc3,       "MBC3" },
  { 0x12, mbc3,       "MBC3+RAM" },
  { 0x13, mbc3,       "MBC3+RAM+BAT" },
  { 0x15, mbcUnknown, "MBC4" },
  { 0x16, mbcUnknown, "MBC4+RAM" },
  { 0x17, mbcUnknown, "MBC4+RAM+BAT" },
  { 0x19, mbc5,       "MBC5" },
  { 0x1a, mbc5,       "MBC5+RAM" },
  { 0x1b, mbc5,       "MBC5+RAM+BAT" },
  { 0x1c, mbc5,       "MBC5+MOT" },
  { 0x1d, mbc5,       "MBC5+RAM+MOT" },
  { 0x1e, mbc5,       "MBC5+RAM+BAT+MOT" },
  { 0xfc, mbcUnknown, "Nintendo Pocket Camera" },
  { 0xfd, mbcUnknown, "Bandai TAMA5" },
  { 0xfe, mbcUnknown, "Hudson HuC-3" },
  { 0xff, mbcUnknown, "Hudson HuC-1" },
  {   -1, mbcUnknown, "Unknown" }
};

/* NOTE:
 * As GB addresses are used as offsets, the addresses for Palm memory
 * blocks are offset by the base addresses.
 */
UInt8 *MemoryRomStart;                    /* address of 1st ROM page */
UInt8 *MemoryRomPages[memoryMaxRomPages]; /* addresses of ROM pages */
static UInt16 RomPage;
UInt8 *MemoryRamPages[memoryMaxRamPages]; /* addresses of RAM pages */
static UInt8 RamPage;

static MemHandle recH[memoryMaxRomPages + 1];
static Boolean   cartridgeHasRam;

UInt8 *MemoryBaseRom0;
UInt8 *MemoryBaseRom1;
UInt8 *MemoryBaseRamVid;
UInt8 *MemoryBaseRamExt;
UInt8 *MemoryBaseRamInt;
UInt8 *MemoryBaseRamOam;

UInt16 MemoryMbc1RomLsb; /* LSByte of MBC1 ROM page */
UInt16 MemoryMbc1RomMsb; /* MSByte of MBC1 ROM page */
UInt8  MemoryMbc1Mode;

UInt16            MemoryMbc3RtcDivider; /* dividing the frame rate */
MemoryMbc3RtcType MemoryMbc3RtcCurrent;
MemoryMbc3RtcType MemoryMbc3RtcLatched;
UInt8             MemoryMbc3Mode;
UInt8             *MemoryMbc3RtcPtr; /* pointer to current register */

typedef struct { /* pointer to functions */
  void *ramEnable;
  void *romSelectL;
  void *romSelectH;
  void *ramSelect;
  void *modeSelect;
} MbcVectorsType;

static const MbcVectorsType DefaultVectors = {
  MemoryMbcDummy,
  MemoryMbcDummy,
  MemoryMbcDummy,
  MemoryMbcDummy,
  MemoryMbcDummy };
const MbcVectorsType Mbc1VectorsMode168 = {
  MemoryMbcDummy,
  MemoryMbc1RomSelectL,
  MemoryMbc1RomSelectL,
  MemoryMbc1RomSelectH,
  MemoryMbc1ModeSelect };
const MbcVectorsType Mbc1VectorsMode432 = {
  MemoryMbcDummy,
  MemoryMbc1RomSelect,
  MemoryMbc1RomSelect,
  MemoryMbc1RamSelect,
  MemoryMbc1ModeSelect };
const MbcVectorsType Mbc2Vectors = {
  MemoryMbcDummy,
  MemoryMbc2RomSelect,
  MemoryMbc2RomSelect,
  MemoryMbcDummy,
  MemoryMbcDummy };
const MbcVectorsType Mbc3Vectors = {
  MemoryMbcDummy,
  MemoryMbc3RomSelect,
  MemoryMbc3RomSelect,
  MemoryMbc3RamSelect,
  MemoryMbc3RtcLatch };
const MbcVectorsType Mbc5Vectors = {
  MemoryMbcDummy,
  MemoryMbc5RomSelect,
  MemoryMbcDummy,
  MemoryMbc5RamSelect,
  MemoryMbcDummy };

/* === Load, save, setup, and cleanup =================================	*/

void MemoryLoadState(MemoryStateType *stateP, Boolean reset) {
  /*--------------------------------------------------------*/
  /* setup pointer to the emulated memories */
  MemoryBaseRamVid = stateP->ram.vid8000 + (0x10000 - 0x8000);
  MemoryBaseRamInt = stateP->ram.intC000 + (0x10000 - 0xc000);
  MemoryBaseRamOam = stateP->ram.oamFe00 + (0x10000 - 0xfe00);

  if (reset) {
    UInt32       t;
    DateType     d;
    DateTimeType dt;

    /* erase memory for a clean screen and some free games */
    MemSemaphoreReserve(true);
    MemSet(&(stateP->ram), sizeof(MemoryRamType), 0x00);
    MemSemaphoreRelease(true);

    RomPage = 0;
    RamPage = 0;

    MemoryMbc1RomLsb = 0;
    MemoryMbc1RomMsb = 0;
    MemoryMbc1Mode = 0;

    MemoryMbc3Mode = 0;
    t = TimGetSeconds();
    DateSecondsToDate(t, &d);
    TimSecondsToDateTime(t, &dt);
    MemoryMbc3RtcCurrent.control = (DateToDays(d) & 0x100) != 0 ?
      0x01 : 0x00;
    MemoryMbc3RtcCurrent.days = (UInt8)DateToDays(d);
    MemoryMbc3RtcCurrent.hours = (UInt8)dt.hour;
    MemoryMbc3RtcCurrent.minutes = (UInt8)dt.minute;
    MemoryMbc3RtcCurrent.seconds = (UInt8)dt.second;

  } else {
    RomPage = stateP->mbc.romPage;
    RamPage = stateP->mbc.ramPage;

    MemoryMbc1RomLsb = stateP->mbc.mbc1RomLsb;
    MemoryMbc1RomMsb = stateP->mbc.mbc1RomMsb;
    MemoryMbc1Mode = stateP->mbc.mbc1Mode;

    MemoryMbc3Mode = stateP->mbc.mbc3Mode;
    MemoryMbc3RtcDivider = stateP->mbc.mbc3RtcDivider;
    MemoryMbc3RtcCurrent = stateP->mbc.mbc3RtcCurrent;
    MemoryMbc3RtcLatched = stateP->mbc.mbc3RtcLatched;
  }
}

void MemorySaveState(void *recP, UInt32 offset) {
  /*-------------------------------------------*/
  MemoryMbcStateType mbc;

  /* the current contents of the emulated is preserved automatically */

  /* look for current ROM page */
  for (mbc.romPage = 0; mbc.romPage < memoryMaxRomPages; mbc.romPage++)
    if (MemoryRomPages[mbc.romPage] == MemoryBaseRom1)
      break;
  mbc.romPage %= memoryMaxRomPages;

  /* look for current ROM page */
  for (mbc.ramPage = 0; mbc.ramPage < memoryMaxRamPages; mbc.ramPage++)
    if (MemoryRamPages[mbc.ramPage] == MemoryBaseRamExt)
      break;
  mbc.ramPage %= memoryMaxRamPages;

  mbc.mbc1RomLsb = MemoryMbc1RomLsb;
  mbc.mbc1RomMsb = MemoryMbc1RomMsb;
  mbc.mbc1Mode = MemoryMbc1Mode;

  mbc.mbc3Mode = MemoryMbc3Mode;
  mbc.mbc3RtcDivider = MemoryMbc3RtcDivider;
  mbc.mbc3RtcCurrent = MemoryMbc3RtcCurrent;
  mbc.mbc3RtcLatched = MemoryMbc3RtcLatched;

  DmWrite(recP, offset + OffsetOf(MemoryStateType, mbc),
	  &mbc, sizeof(MemoryMbcStateType));
}

Boolean MemorySetup(DmOpenRef dbP) {
  /*------------------------------*/
  UInt16               numRecs, i, romPages;
  mbcType              mbc;
  UInt8                *ramExtPtr;
  UInt16               ramPageOffset;
  const MbcVectorsType *mbcVectorsP;

  /* get number of records and check it */
  numRecs = DmNumRecords(dbP);
  if (numRecs == 0 || numRecs > memoryMaxRomPages)
    return false;

  /* lock all records */
  for (i = 0; i < numRecs; i++)
    recH[i] = DmGetRecord(dbP, i);

  /* save whether cartridge has RAM */
  cartridgeHasRam = (numRecs & 1) == 0;

  /* setup ROM pages */
  romPages = numRecs + 1 - (cartridgeHasRam ? 1 : 0);
  MemoryRomStart = MemHandleLock(recH[0]);
  MemoryRomPages[0] = MemoryRomStart + 0x0000 - 0x4000;
  MemoryRomPages[1] = MemoryRomStart + 0x4000 - 0x4000;
  for (i = 2; i < romPages; i++)
    MemoryRomPages[i] = MemHandleLock(recH[i - 1]) - 0x4000;

  /* setup cartridge RAM */
  if (cartridgeHasRam)
    ramExtPtr = MemHandleLock(recH[numRecs - 1]) + (0x10000 - 0xa000);
  else
    ramExtPtr = NULL;

  /* Wenn die Patches implementiert werden, muss zwar die CRC ueber den
   * ungepatchten Inhalt erzeugt werden. Aber bevor hier die weiteren
   * Entscheidungen getroffen werden, werden die Patches angebracht. Die
   * geaenderten Daten liegen dann auch im RAM... Hm, das wird nicht ganz
   * trivial zu implementieren.
   */

  /* check for correct combination of mbc and ram size */
  for (i = 0; MemoryCartridgesKnown[i].code != -1; i++)
    if (MemoryRomStart[0x147] == MemoryCartridgesKnown[i].code)
      break;
  mbc = MemoryCartridgesKnown[i].mbc;
  if (mbc == mbcUnknown)
    return false;
  if (cartridgeHasRam) {
    if ((MemoryRomStart[0x149] == 0 && mbc != mbc2) ||
	MemoryRomStart[0x149] > 3)
      return false;
  } else {
    if (MemoryRomStart[0x149] != 0 || mbc == mbc2)
      return false;
  }

  /* setup RAM access vectors and pages */
  MemSemaphoreReserve(true);

  if (cartridgeHasRam) {
    switch (MemoryRomStart[0x149]) {
    case 1:
      EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address =
	MemoryRamRdExt2K;
      EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address =
	MemoryRamWrExt2K;
      ramPageOffset = 0x0000;
      break;
    case 2:
      EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address =
	MemoryRamRdExt8K;
      EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address =
	MemoryRamWrExt8K;
      ramPageOffset = 0x0000;
      break;
    case 3:
      EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address =
	MemoryRamRdExt8K;
      EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address =
	MemoryRamWrExt8K;
      ramPageOffset = 0x2000;
      break;
    default: /* 0, for mbc2 */
      EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address =
	MemoryRamRdExtMbc2;
      EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address =
	MemoryRamWrExtMbc2;
      ramPageOffset = 0x0000;
      break;
    }
  } else {
    EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address =
      MemoryRamRdExtNone;
    EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address =
      MemoryRamWrExtNone;
    ramPageOffset = 0x0000;
  }
  MemoryRamPages[0] = ramExtPtr + 0 * ramPageOffset;
  MemoryRamPages[1] = ramExtPtr + 1 * ramPageOffset;
  MemoryRamPages[2] = ramExtPtr + 2 * ramPageOffset;
  MemoryRamPages[3] = ramExtPtr + 3 * ramPageOffset;

  /* setup MBC vectors */
  switch (mbc) {
  case mbc1:
    for (i = 0; i < romPages; i += 32)
      MemoryRomPages[i] = MemoryRomPages[i + 1];
    if ((MemoryMbc1Mode & 0x01) == 0)
      mbcVectorsP = &Mbc1VectorsMode168;
    else
      mbcVectorsP = &Mbc1VectorsMode432;
    break;
  case mbc2:
    MemoryRomPages[0] = MemoryRomPages[1];
    mbcVectorsP = &Mbc2Vectors;
    break;
  case mbc3:
    EmulationEmulator[gbemuJmpRamRdExt].xt.vector.buffer =
      EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address;
    EmulationEmulator[gbemuJmpRamWrExt].xt.vector.buffer =
      EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address;
    if ((MemoryMbc3Mode & 0x08) != 0) {
      EmulationEmulator[gbemuJmpRamRdExt].xt.vector.address =
	MemoryMbc3RtcRd;
      EmulationEmulator[gbemuJmpRamWrExt].xt.vector.address =
	MemoryMbc3RtcWr;
      MemoryMbc3RtcPtr = (UInt8 *)&MemoryMbc3RtcLatched +
	(MemoryMbc3Mode & 0x07);
    }
    MemoryRomPages[0] = MemoryRomPages[1];
    mbcVectorsP = &Mbc3Vectors;
    break;
  case mbc5:
    mbcVectorsP = &Mbc5Vectors;
    break;
  default:
    MemoryRomPages[0] = MemoryRomPages[1];
    mbcVectorsP = &DefaultVectors;
    break;
  }
  EmulationEmulator[gbemuJmpRamEnable ].xt.vector.address =
    mbcVectorsP->ramEnable;
  EmulationEmulator[gbemuJmpRomSelectL].xt.vector.address =
    mbcVectorsP->romSelectL;
  EmulationEmulator[gbemuJmpRomSelectH].xt.vector.address =
    mbcVectorsP->romSelectH;
  EmulationEmulator[gbemuJmpRamSelect ].xt.vector.address =
    mbcVectorsP->ramSelect;
  EmulationEmulator[gbemuJmpModeSelect].xt.vector.address =
    mbcVectorsP->modeSelect;
  for (i = romPages; i < memoryMaxRomPages; i++)
    MemoryRomPages[i] = MemoryRomPages[i - romPages];

  /* restore current ROM and RAM page */
  MemoryBaseRom1   = MemoryRomPages[RomPage];
  MemoryBaseRamExt = MemoryRamPages[RamPage];

  MemSemaphoreRelease(true);
  return true;
}

void MemoryCleanup(DmOpenRef dbP) {
  /*-----------------------------*/
  UInt16 numRecs, i;

  numRecs = DmNumRecords(dbP);

  for (i = 0; i < numRecs - 1; i++) {
    MemHandleUnlock(recH[i]);
    DmReleaseRecord(dbP, i, false);
  }

  /* the last record is marked dirty, if cartridge has RAM */
  MemHandleUnlock(recH[i]);
  DmReleaseRecord(dbP, i, cartridgeHasRam);
}

/* === The end ========================================================	*/
