/************************************************************************
 *  palmboy.c
 *                      Nintendo Gameboy(TM) Emulator for
 *                      Palm OS(R) Computing Platform
 *
 *  NOTES:
 *  "Palm" and "Palm OS" are registered trademarks of Palm.
 *  "Gameboy" is a registered trademark of Nintendo.
 *
 *  2000-03-28  Bodo Wenzel     v.0.0, Creation of user interface without
 *                              emulator mock-up and gray image
 *  2000-04-15  Bodo Wenzel     v.0.1, Creation of user interface without
 *                              gray image
 *  2000-04-19  Bodo Wenzel     v.0.2, Creation of user interface
 *  2000-04-25  Bodo Wenzel     v.0.3, Errors fixed with Palm IIIC, V
 *  2000-05-01  Bodo Wenzel     v.0.4, Quick'n'dirty screen conversion,
 *                              splitting in modules, and usage of more
 *                              registers
 *  2000-05-06  Bodo Wenzel     v.0.9, First working version with built-in
 *                              games
 *  2000-05-07  Bodo Wenzel     v.1.0, First working version
 *  2000-09-11  Bodo Wenzel     v.2.0, Nearly full screen emulation
 *  2000-09-30  Bodo Wenzel     v.2.1, Scroller in GB code
 *  2000-10-16  Bodo Wenzel     v.3.0, Interrupts and timer
 *  2000-11-03  Bodo Wenzel     v.3.0a, No more "Visor crash"
 *  2000-11-18  Bodo Wenzel     v.3.1, Support of color devices
 *  2001-02-22  Bodo Wenzel     v.3.2 (v.3.14159), Hsync interrupts and
 *                              support of more cartridges
 *  2001-06-23  Bodo Wenzel     v.3.3, Bug in MBC1 code corrected
 *  2001-07-01  Bodo Wenzel     v.3.3b, Bugfix corrected ;-)
 *  2001-07-23  Bodo Wenzel     v.4.0, Many sources converted to C,
 *                              compiler/assembler now GCC
 ************************************************************************
 *
 * (c)2000-2001 Bodo Wenzel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

/* IMPORTANT!
 * Please make sure that any referenced resource is included!
 * Unfortunately there is no automatic!
 */

/* === Includes =======================================================	*/

/* PalmBoy needs OS3.5 support for grayscale */
#include <PalmOS.h>

#if 0 /* GamePad, kommt erst mit UI6 integration */
/* GamePad support */
#define UInt8 UInt8 /* they check it with #ifdef :-( */
#include <GPDLib.h> /* war lokal, root sollte den link setzen */
#endif

#include "palmboy.h"
#include "list.h"
#include "emulation.h"
#include "video.h"

/* === Local function prototypes ======================================	*/

static Err StartApplication(UInt16 launchFlags);
static void StopApplication(void);
static void EventLoop(void);
static Boolean ApplicationHandleEvent(EventType *evtP);

static Err RomVersionCompatible(UInt32 requiredV, Int16 launchFlags);

/* === M A I N ========================================================	*/

UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags) {
  /*----------------------------------------------------------*/
  Err ret;

  ret = errNone;

  if (cmd == sysAppLaunchCmdNormalLaunch) {
    /* normal launch */

    ret = StartApplication(launchFlags);
    if (ret == errNone) {
      if (EmulationRuns)
	FrmGotoForm(formIdEmulation);
      else
	FrmGotoForm(formIdList);

      EventLoop();
      StopApplication();
    }
  }

  return ret;
}

/* === Local functions ================================================	*/

static Err StartApplication(UInt16 launchFlags) {
  /*-------------------------------------------*/
  Err err;

  err = RomVersionCompatible(
    sysMakeROMVersion(3, 0, 0, sysROMStageRelease, 0), launchFlags);
  if (err != errNone)
    return err;

  if (!EmulationInit()) {
    FrmAlert(alertIdNoEmulator);
    return dmErrResourceNotFound;
  }

  if (!VideoGraySwitchOn()) {
    FrmAlert(alertIdIncompatibleScreen);
    return sysErrParamErr;
  }

#if 0 /* prefs init */
  PrefsInit();
#endif
  ListInit();
#if 0 /* patch init */
  PatchInit();
#endif
  return errNone;
}

static void StopApplication(void) {
  /*-----------------------------*/
  FrmCloseAllForms();

#if 0 /* patch exit */
  PatchExit();
#endif
  ListExit();
#if 0 /* prefs exit */
  PrefsExit();
#endif

  VideoGraySwitchOff();

  EmulationExit();
}

static void EventLoop(void) {
  /*-----------------------*/
#if 1 /* bis ui6 hier TheEvent */
  EventType TheEvent;
#endif
  Err err;

  do {
#if 0 /* noch nicht ui6-GetEvent() */
    EmulationGetEvent();
#else
    EvtGetEvent(&TheEvent, evtWaitForever);
#endif
    if (!SysHandleEvent(&TheEvent))
      if (!MenuHandleEvent(NULL, &TheEvent, &err))
	if (!ApplicationHandleEvent(&TheEvent))
	  FrmDispatchEvent(&TheEvent);
  } while (TheEvent.eType != appStopEvent);
}

static Boolean ApplicationHandleEvent(EventType *evtP) {
  /*--------------------------------------------------*/
  FormType *frmP;

  if (evtP->eType == frmLoadEvent) {
    frmP = FrmInitForm(evtP->data.frmLoad.formID);
    FrmSetActiveForm(frmP);

    switch (evtP->data.frmLoad.formID) {
    case formIdList:
      FrmSetEventHandler(frmP, ListFormHandleEvent);
      break;
    case formIdEmulation:
      FrmSetEventHandler(frmP, EmulationFormHandleEvent);
      break;
    default:
      break;
    }
    return true;
  }
  return false;
}

static Err RomVersionCompatible(UInt32 requiredV, Int16 launchFlags) {
  /*----------------------------------------------------------------*/
  /* Compare the actual ROM version with the required one. If it's not
   * good enough, return with an error code. But if it's less than OS
   * version 2.0 start the default application - otherwise this app will
   * be started over and over again!
   */
  UInt32 romV;

  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romV);
  if (romV >= requiredV)
    return errNone;

  if ((launchFlags &
       (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
      (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) {
    FrmAlert(alertIdIncompatibleRom);
    if (romV < sysMakeROMVersion(2, 0, 0, sysROMStageDevelopment, 0))
      AppLaunchWithCommand(sysFileCDefaultApp, sysAppLaunchCmdNormalLaunch,
			   NULL);
  }

  return sysErrRomIncompatible;
}

/* === The end ========================================================	*/
