/************************************************************************
 *  video_c.c
 *                      Video emulation routines for PalmBoy
 *                      (C functions)
 *
 *  emuvideo.asm:
 *  2000-09-10  Bodo Wenzel   Got from runmode.asm
 *  2000-09-17  Bodo Wenzel   Now nearly full screen emulation
 *  2000-10-23  Bodo Wenzel   Interrupts and timer
 *  2000-11-18  Bodo Wenzel   Support of color devices
 *  2001-02-18  Bodo Wenzel   Window bug removed,
 *                            bug with visibility of objects found,
 *                            interrupts for hsync implemented
 *  gray.asm:
 *  2000-05-05  Bodo Wenzel   Got from full source v.0.4
 *  2000-07-26  Bodo Wenzel   Longer off time while mode switching
 *  2000-09-13  Bodo Wenzel   Now nearly full screen emulation
 *  2000-11-01  Till Harbaum  Added support for the Palm IIIc and friends
 *  2000-11-19  Bodo Wenzel   Clean-up of support for color devices
 *  2001-02-10  Bodo Wenzel   Support for grayhack: no register writes
 *
 *  2001-07-22  Bodo Wenzel   Separated from emuvideo.asm, joined with
 *                            parts of gray.asm, changed to C
 *  2001-09-16  Bodo Wenzel   Saving states now possible
 ************************************************************************
 *
 * (c)2000-2001 Bodo Wenzel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 ************************************************************************
 */

/* === Includes =======================================================	*/

#include <PalmOS.h>

#include "palmboy.h"
#include "video.h"
#include "emulation.h"
#include "memory.h"
#include "gbemu.h"

/* === Constants ======================================================	*/

#define bytesPerBuffer (videoScreenHeight * \
                        videoScreenWidth / 8 * videoBitsPerPixel)

/* === Global variables ===============================================	*/

UInt16 VideoCycles; /* opcodes until next lcd line */

UInt16 VideoTimerDiv;       /* fix-real for DIV */
UInt16 VideoTimerTima;      /* fix-real for TIMA */
UInt16 VideoTimerDeltaTima; /* increment for TIMA */

UInt16 *VideoMapIndices = NULL; /* pointer to allocated MapIndices */

/* both of pattern and mask are
 *   8 pixel lines in
 *     $100 possible tiles of
 *       1 shifted word as long
 */
typedef struct {
  UInt8  used[0x100];       /* which tiles are used */
  UInt32 tile[2][8][0x100]; /* converted tiles */
} VideoBgWiRamType;

static VideoBgWiRamType *BgRamPtr = NULL;
void *VideoBgUsed;
void *VideoBgTile;

static VideoBgWiRamType *WiRamPtr = NULL;
void *VideoWiUsed;
void *VideoWiTile;

/* (SCR_HEIGHT*8) lines with
 *   1 word offset to first background object
 *   1 word offset to first foreground object
 * 40 objects of
 *   16 pixel lines with
 *     1 word screen horizontal offset
 *     1 shifted word as long mask
 *     1 shifted word as long pattern
 *     1 word next offset
 *     1 word prev offset
 */
typedef struct {
  UInt16 dummy; /* for offset -2 */
  struct {
    UInt16 bg;
    UInt16 fg;
  } first[videoScreenHeight]; /* offsets */
  struct {
    UInt16 hOffset;
    UInt32 mask;
    UInt32 pattern;
    UInt16 next;
    UInt16 prev;
  } tiles[40][16]; /* converted tiles */
} VideoObRamType;

static VideoObRamType *ObRamPtr = NULL;
void *VideoObFirst;
void *VideoObTiles;

void *VideoBufferPtr = NULL;

void *VideoDisplayPtr;    /* pointer to display memory */
UInt16 VideoDisplayWidth; /* offset between two pixel lines */

UInt8 VideoSavedLcdc;

/* === Function prototypes ============================================	*/

static void Compress(UInt8 *destP);
static void Expand(UInt8 *srceP);

/* === Functions for switching the screen mode ========================	*/

Boolean VideoGraySwitchOn(void) {
  /*---------------------------*/
  UInt32     depth;
  UInt32     romV;
  WinHandle  winH;

  /* set display into 4-level-gray mode */
  depth = videoBitsPerPixel;
  if (WinScreenMode(winScreenModeSet, NULL, NULL, &depth, NULL) != errNone)
    return false;

  /* get important display values */
  winH = WinGetDisplayWindow();
  FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romV);
  if (romV >= sysMakeROMVersion(3, 5, 0, sysROMStageDevelopment, 0)) {
    /* hier fehlt noch die Umschaltung fuer die HandEra H330 */
    BitmapType *bmpP;

    bmpP = WinGetBitmap(winH);
    VideoDisplayPtr = BmpGetBits(bmpP);
    VideoDisplayWidth = bmpP->rowBytes;

  } else {
    WindowType *winP;

    winP = WinGetWindowPointer(winH);
    VideoDisplayPtr = winP->displayAddrV20;
    VideoDisplayWidth = winP->displayWidthV20 / (8 / videoBitsPerPixel);
  }
  return true;
}

void VideoGraySwitchOff(void) {
  /*-------------------------*/
  WinScreenMode(winScreenModeSetToDefaults, NULL, NULL, NULL, NULL);
}

/* === Load, save, setup and cleanup ==================================	*/

void VideoLoadState(VideoStateType *stateP, Boolean reset) {
  /*------------------------------------------------------*/
  if (reset) {
    EmulationIo.bgp =  0xfc;
    EmulationIo.obp0 = 0xff;
    EmulationIo.obp1 = 0xff;

  } else {
    VideoTimerDiv = stateP->timerDiv;
    VideoTimerTima = stateP->timerTima;
  }
}

void VideoSaveState(void *recP, UInt32 offset) {
  /*------------------------------------------*/
  static const BitmapType bmp = {
    videoScreenWidth / 2, videoScreenHeight / 2,
    videoScreenWidth / 2 / 8 * videoBitsPerPixel,
    { 0, 0, 0, 0, 0, 0 },
    videoBitsPerPixel, 1, 0, 0, BitmapCompressionTypeNone, 0
  };

  DmWrite(recP, offset + OffsetOf(VideoStateType, timerDiv),
	  &VideoTimerDiv, sizeof(VideoTimerDiv));
  DmWrite(recP, offset + OffsetOf(VideoStateType, timerTima),
	  &VideoTimerTima, sizeof(VideoTimerTima));

  DmWrite(recP, offset + OffsetOf(VideoStateType, bitmap.bmp),
	  &bmp, sizeof(BitmapType));
  Compress(recP + offset + OffsetOf(VideoStateType, bitmap.data));
}

Boolean VideoSetup(DmOpenRef dbP, VideoStateType *stateP) {
  /*-----------------------------------------------------*/
  UInt16    recId;
  MemHandle recH;

  /* get additional memory */

  recId = dmMaxRecordIndex;
  recH = DmNewRecord(dbP, &recId,
		     2 * videoMapWrap * videoMapWrap * videoBitsPerPixel);
  if (recH == NULL)
    return false;

  VideoMapIndices = MemHandleLock(recH);
  if (VideoMapIndices == NULL)
    return false;

  recId = dmMaxRecordIndex;
  recH = DmNewRecord(dbP, &recId, sizeof(VideoBgWiRamType));
  if (recH == NULL)
    return false;

  BgRamPtr = MemHandleLock(recH);
  if (BgRamPtr == NULL)
    return false;
  VideoBgUsed = BgRamPtr->used;
  VideoBgTile = BgRamPtr->tile;

  recId = dmMaxRecordIndex;
  recH = DmNewRecord(dbP, &recId, sizeof(VideoBgWiRamType));
  if (recH == NULL)
    return false;

  WiRamPtr = MemHandleLock(recH);
  if (WiRamPtr == NULL)
    return false;
  VideoWiUsed = WiRamPtr->used;
  VideoWiTile = WiRamPtr->tile;

  recId = dmMaxRecordIndex;
  recH = DmNewRecord(dbP, &recId, sizeof(VideoObRamType));
  if (recH == NULL)
    return false;

  ObRamPtr = MemHandleLock(recH);
  if (ObRamPtr == NULL)
    return false;
  VideoObFirst = ObRamPtr->first;
  VideoObTiles = ObRamPtr->tiles;

  recId = dmMaxRecordIndex;
  recH = DmNewRecord(dbP, &recId, bytesPerBuffer);
  if (recH == NULL)
    return false;

  VideoBufferPtr = MemHandleLock(recH);
  if (VideoBufferPtr == NULL)
    return false;

  /* setup some vectors */

  MemSemaphoreReserve(true);

  EmulationEmulator[0x07].out.vector.address = VideoIoWrTac;

  EmulationEmulator[0x40].out.vector.address = VideoIoWrLcdc;
  EmulationEmulator[0x41].in.vector.address = VideoIoRdStat;
  EmulationEmulator[0x41].out.vector.address = VideoIoWrStat;
  EmulationEmulator[0x42].out.vector.address = VideoIoWrScy;
  EmulationEmulator[0x43].out.vector.address = VideoIoWrScx;
  EmulationEmulator[0x45].out.vector.address = VideoIoWrLyc;
  EmulationEmulator[gbemuJmpIoWrDma ].xt.vector.address = VideoIoWrDma;
  EmulationEmulator[0x47].out.vector.address = VideoIoWrBgp;
  EmulationEmulator[0x48].out.vector.address = VideoIoWrObp0;
  EmulationEmulator[0x49].out.vector.address = VideoIoWrObp1;
  EmulationEmulator[0x4a].out.vector.address = VideoIoWrWy;
  EmulationEmulator[0x4b].out.vector.address = VideoIoWrWx;

  /* Some memory and variables have to be changed and will be corrected by
   * the setup code. This ensures that all video variables are set up
   * properly and the pre-calculated variables don't have to be saved with
   * the game's state.
   * The LCDC register has to be saved because it can't be inverted simply.
   */
  VideoSavedLcdc = EmulationIo.lcdc;
  EmulationIo.lcdc = 0x80;
  EmulationIo.bgp = ~EmulationIo.bgp;
  EmulationIo.obp0 = ~EmulationIo.obp0;
  EmulationIo.obp1 = ~EmulationIo.obp1;
  {
    UInt8  *p;
    UInt16 i;
    for (p = MemoryBaseRamVid - (0x10000 - 0x8000), i = 0x1800;
	 i < 0x2000; i++)
      p[i] = ~p[i];
  }

  /* create default screen contents */
  if (EmulationRuns)
    Expand(stateP->bitmap.data);
  else
    MemSet(VideoBufferPtr, bytesPerBuffer, 0);

  MemSemaphoreRelease(true);

  return true;
}

void VideoCleanup(void) {
  /*-------------------*/
  if (VideoMapIndices != NULL) {
    MemPtrUnlock(VideoMapIndices);
    VideoMapIndices = NULL;
  }

  if (BgRamPtr != NULL) {
    MemPtrUnlock(BgRamPtr);
    BgRamPtr = NULL;
  }

  if (WiRamPtr != NULL) {
    MemPtrUnlock(WiRamPtr);
    WiRamPtr = NULL;
  }

  if (ObRamPtr != NULL) {
    MemPtrUnlock(ObRamPtr);
    ObRamPtr = NULL;
  }

  if (VideoBufferPtr != NULL) {
    MemPtrUnlock(VideoBufferPtr);
    VideoBufferPtr = NULL;
  }
}

/* === Functions for the screen thumbnail =============================	*/

static void Compress(UInt8 *destP) {
  /*------------------------------*/
  UInt8  *src1P, *src2P, w;
  UInt16 x, y;

  MemSemaphoreReserve(true);

  src1P = VideoDisplayPtr;

  for (y = 0; y < videoScreenHeight / 2; y++) {
    src2P = src1P + VideoDisplayWidth;
    for (x = 0; x < videoScreenWidth / 2 / (8 / videoBitsPerPixel); x++) {
      w  = ((*src1P & 0xC0) + (*src2P & 0xC0)) >> 2;
      w +=  (*src1P & 0x30) + (*src2P & 0x30);
      w += 0x20;
      *destP = w & 0xC0;
      w  =  (*src1P & 0x0C) + (*src2P & 0x0C);
      w += ((*src1P & 0x03) + (*src2P & 0x03)) << 2;
      w += 0x08;
      *destP |= w & 0x30;
      src1P++;
      src2P++;
      w  = ((*src1P & 0xC0) + (*src2P & 0xC0)) >> 2;
      w +=  (*src1P & 0x30) + (*src2P & 0x30);
      w += 0x20;
      *destP |= (w >> 4) & 0x0C;
      w  =  (*src1P & 0x0C) + (*src2P & 0x0C);
      w += ((*src1P & 0x03) + (*src2P & 0x03)) << 2;
      w += 0x08;
      *destP |= (w >> 4) & 0x03;
      src1P++;
      src2P++;
      destP++;
    }
    src1P = src2P + VideoDisplayWidth -
      videoScreenWidth / (8 / videoBitsPerPixel);
  }

  MemSemaphoreRelease(true);
}

static void Expand(UInt8 *srceP) {
  /*----------------------------*/
  UInt8  *destP;
  UInt16 x, y;

  destP = VideoBufferPtr;

  for (y = 0; y < videoScreenHeight / 2; y++) {
    for (x = 0; x < videoScreenWidth / 2 / (8 / videoBitsPerPixel); x++) {
      UInt8 pixel;

      pixel  = (*srceP & 0xC0) | ((*srceP & 0x30) >> 2);
      pixel |= pixel >> 2;
      *(destP + videoScreenWidth / (8 / videoBitsPerPixel)) = pixel;
      *destP++ = pixel;
      pixel  = ((*srceP & 0x0C) << 2) | (*srceP & 0x03);
      pixel |= pixel << 2;
      *(destP + videoScreenWidth / (8 / videoBitsPerPixel)) = pixel;
      *destP++ = pixel;
      srceP++;
    }
    destP += videoScreenWidth / (8 / videoBitsPerPixel);
  }
}

/* === The end ========================================================	*/
